/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <string.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"

extern int	nextAtomicId;
extern int	debugMode;
extern char*	types[NumTypes];

extern BAIdList* private_cp;
extern BAIdList* shared_cp;

ooHandle(ooContObj) DocContainerH;

////////////////////////////////////////////////////////////////////////////
//
// CompositePart Constructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::CompositePart(int32 cpId)
{
    ooHandle(BaseAssembly) baH;
    ooHandle(CompositePart) thisH;
    thisH = ooThis();

    if (debugMode) {
        printf("CompositePart::CompositePart(cpId = %d)\n", cpId);
    }

    // initialize the simple stuff (some of it randomly)

    id = cpId;
    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);

    // for the build date, decide if this part is young or old, and then
    // randomly choose a date in the required range

    if (cpId % YoungCompFrac == 0) {
        // young one
        if (debugMode) {
            fprintf(stderr, "(young composite part, id = %d.)\n", id);
        }
        buildDate = MinYoungCompDate + 
		    (int)(random() % (MaxYoungCompDate - MinYoungCompDate + 1));
    } else { 
        // old one
        if (debugMode) {
            fprintf(stderr, "(old composite part, id = %d.)\n", id);
        }
        buildDate = MinOldCompDate + 
		    (int)(random() % (MaxOldCompDate - MinOldCompDate + 1));
    }

    // initialize the documentation (indexed by its title) ...
    // WARNING:  The actual title is actually inside the Document
    // constructor, but doing that isn't possible in this version
    // of Objectivity/DB, so it's also determined out here.  (To
    // create an indexed object, it appears that the key has to
    // be determined before the constructor call, or at least that
    // the place where it's value will go has to be predetermined.)

    ooKey key;
    key.type = oocString;
    key.offset = ooGetMemberOffset(Document, title);
    key.size   = ooGetMemberSize(Document, title);
    char title[TitleSize];
    sprintf(title, "Composite Part %08d", cpId);
    key.value  = title;

    ooHandle(CompositePart) compH = ooThis();
    ooHandle(Document) docH = ooNewKey(Document, 
                                       (cpId), 
                                       DocContainerH, key);
    set_documentation(docH);

    // now create the atomic parts (indexed by their ids) ...

    int i, id, from, to;
    ooHandle(AtomicPart)* atomH = new ooHandle(AtomicPart) [NumAtomicPerComp];
    ooHandle(Connection) connH;

    key.type = oocInt32;
    key.offset = ooGetMemberOffset(AtomicPart, id);
    key.size   = ooGetMemberSize(AtomicPart, id);
    key.value  = &id;

    for (i = 0; i < NumAtomicPerComp; i++) {
	id = nextAtomicId + i;
	atomH[i] = ooNewKey(AtomicPart, (id), compH, key);
	add_parts(atomH[i]);
	if (i == 0) { set_rootPart(atomH[i]); }
    }

    // ... and then wire them semi-randomly together (as a ring plus random
    // additional connections to ensure full part reachability for traversals)

    for (from = 0; from < NumAtomicPerComp; from++) {
	for (i = 0; i < NumConnPerAtomic; i++) {
	    if (i == 0) {
	        to = (from + 1) % NumAtomicPerComp;
	    } else {
	        to = (int) (random() % NumAtomicPerComp);
	    }
	    connH = new(compH) Connection(atomH[from], atomH[to]);
	}
    }
    delete [NumAtomicPerComp] atomH;
    nextAtomicId += NumAtomicPerComp;

    // finally insert this composite part as a child of the base
    // assemblies that use it

    // first the assemblies using the comp part as a shared component
    shared_cp[cpId].initScan();

    // get the first base assembly
    baH = shared_cp[cpId].next();
    while (!baH.isNull())
    {
        // add this composite part cp the list of shared parts used
        // in the assembly
	baH->add_componentsShar(thisH);

        // continue with next base assembly
        baH = shared_cp[cpId].next();
    }

    // next the assemblies using the comp part as a private component
    private_cp[cpId].initScan();

    // get the first base assembly
    baH = private_cp[cpId].next();
    while (!baH.isNull())
    {
        // add this composite part to the list of private parts used
        // in the assembly
	baH->add_componentsPriv(thisH);

        // continue with next base assembly
        baH = private_cp[cpId].next();
    }
}


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Destructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::~CompositePart()
{
    ooHandle(Document) docH;

    if (debugMode) {
        printf("CompositePart::~CompositePart(cpId = %d)\n", id);
    }

    // explicitly delete document and atomic parts (workaround due to
    // prop(delete) being broken at present for inline associations)

    docH = documentation();
    if (!ooDeleteNoProp(docH)) {
	fprintf(stderr, "ERROR: Unable to delete document\n");
    }

    ooItr(AtomicPart) atomI;
    for (parts(atomI); atomI.next(); ) {
	if (!ooDeleteNoProp(atomI)) {
	    fprintf(stderr, "ERROR: Unable to delete an atomic part\n");
	}
    }
}

