/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


//--------------------------------------------------------------------------
// Start with some necessary preliminaries
//--------------------------------------------------------------------------

declare (ooVArray, char);

#define TypeSize	10
#define TitleSize	40
#define DummySize	1000

const int FALSE = 0;
const int TRUE  = 1;

typedef enum { Trav1, Trav2a, Trav2b, Trav2c, Trav3a, Trav3b, Trav3c, Trav4,
	       Trav5do, Trav5undo, Trav6, Trav7, Trav8, Trav9, Trav10,
	       Trav1WW,
               Query1, Query1WW, Query2, Query3, Query4,
	       Query5, Query6, Query7, Query8, Insert, Delete, Reorg1, Reorg2,
	       WarmUpdate, MultiTrav1, MultiTrav2, MultiTrav3, MultiTrav4, 
	       MultiTrav5, MultiTrav6 } BenchmarkOp;

#include "PartIdSet.h"

//--------------------------------------------------------------------------
// AtomicPart objects are the primitives for building up designs
//	- modeled after the Sun/OO1 benchmark's parts
//--------------------------------------------------------------------------

class AtomicPart: public ooObj {
public:
    int32			id;
    char			type[TypeSize];
    int32			buildDate;
    int32			x, y;
    int32                       docId;
    ooHandle(Connection)	to[] <-> from : inline(short);
    ooHandle(Connection)	from[] <-> to : inline(short);
    ooHandle(CompositePart)	partOf <-> parts[] : inline(short);

    AtomicPart(int32 ptId);
    ~AtomicPart();
    void swapXY();
    void toggleDate();
    int traverse(BenchmarkOp op, PartIdSet& visitedIds);
    void DoNothing();
};


//--------------------------------------------------------------------------
// Connection objects are used to wire AtomicParts together
//	- similarly, modeled after Sun/OO1 connections
//--------------------------------------------------------------------------

class Connection: public ooObj {
public:
    char			type[TypeSize];
    int32			length;
    ooHandle(AtomicPart)	from <-> to[] : inline(short);
    ooHandle(AtomicPart)	to <-> from[] : inline(short);

    Connection(ooHandle(AtomicPart)& fromPart, ooHandle(AtomicPart)& toPart);
};


//--------------------------------------------------------------------------
// CompositeParts are parts constructed from AtomicParts
//	- entry in a library of reusesable components
//	- implementation is a graph of atomic parts
//	- provides unit of significant access locality
//	- each has an associated (unique) document object
//--------------------------------------------------------------------------

class CompositePart: public ooObj {
public:
    int32			id;
    char			type[TypeSize];
    int32			buildDate;
    ooHandle(Document)		documentation <-> part : inline;
    ooHandle(BaseAssembly)	usedInPriv[] <-> componentsPriv[] : inline;
    ooHandle(BaseAssembly)	usedInShar[] <-> componentsShar[] : inline;
    ooHandle(AtomicPart)	parts[] <-> partOf : inline(short);
    ooHandle(AtomicPart)	rootPart : inline(short);

    CompositePart(int32 cpId);
    ~CompositePart();
    int traverse(BenchmarkOp op);
    int traverse7();
    int reorg1(ooHandle(ooContObj)& designLibH);
    int reorg2(ooHandle(ooContObj)& designLibH);
};


//--------------------------------------------------------------------------
// Document objects are used to hold a description of some particular
// CompositePart object
//--------------------------------------------------------------------------

class Document: public ooObj {
public:
    char			title[TitleSize];
    int32                       id;
    ooVArray(char)		text;
    ooHandle(CompositePart)	part <-> documentation : inline;

    Document(int32 cpId);
    int searchText(char c);
    int replaceText(char *oldString, char* newString);
};


//--------------------------------------------------------------------------
// Manual objects are used to hold a description of some particular
// module.  They are identical to very large documents; they are declared as
// a separate type to make things easier in systems that maintain
// type extents (we don't want to put manuals and documents in the
// same extent --- for one thing, id's are not unique between
// the two.)  Also, the bidirectional association declarations would
// require (1) both Modules and CompositeParts to refer to these things
// by the same name, and (2) Documents to refer to the thing they
// document (both modules and composite parts) by the same name.  Even
// worse, it would be hard to make the (typed) handle in the document
// refer to both modules and composite parts without messing things
// up.
//--------------------------------------------------------------------------

class Manual: public ooObj {
public:
    char			title[TitleSize];
    int32                       id;
    ooVArray(char)		text;
    int                         textLen;  // length of "text".
    ooHandle(Module)	mod <-> man : inline(short);

    Manual(int32 manId);
    int searchText(char c);
    int firstLast();
};



//--------------------------------------------------------------------------
// Assembly objects are design instances built up recursively from
// from other Assembly objects and (at the leaves only) CompositeParts
//	- hierarchical (tree) structure for designs
//	- may share composite parts with other assemblies
//	- nonleaf and leaf assembly subtypes
//--------------------------------------------------------------------------

class Assembly: public ooObj {
public:
    int32			id;
    char			type[TypeSize];
    int32			buildDate;
    ooHandle(ComplexAssembly)	superAssembly <-> subAssemblies[]
    							: inline(short);

    virtual int traverse(BenchmarkOp op) = 0;
    virtual int traverse7(PartIdSet& visitedComplexIds) = 0;
    void DoNothing();
};


class ComplexAssembly: public Assembly {
public:
    ooHandle(Assembly)		subAssemblies[] <-> superAssembly
   							 : inline(short);

    ComplexAssembly(int32 asId, ooHandle(ComplexAssembly)& parentAssembly,
    		    int32 levelNo);
    int traverse(BenchmarkOp op);
    int traverse7(PartIdSet& visitedComplexIds);
};


class BaseAssembly: public Assembly {
public:
    ooHandle(CompositePart)	componentsPriv[] <-> usedInPriv[] : inline;
    ooHandle(CompositePart)	componentsShar[] <-> usedInShar[] : inline;

    BaseAssembly(int32 asId, ooHandle(ComplexAssembly)& parentAssembly);
    int traverse(BenchmarkOp op);
    int traverse7(PartIdSet& visitedComplexIds);
};


//--------------------------------------------------------------------------
// Modules are the designs resulting from Assembly composition
//	- unit of scaleup for the benchmark database
//	- may share composite parts with other modules
//--------------------------------------------------------------------------

class Module: public ooContObj {
public:
    int32			id;
    char			type[TypeSize];
    int32			buildDate;
    ooHandle(ComplexAssembly)	designRoot : inline;
    ooHandle(Manual)		man <-> mod : inline(short);

    Module(int32 modId);
    int traverse(BenchmarkOp op);
    int scanManual();
    int firstLast();
};


//--------------------------------------------------------------------------
// Dummy objects -- used for cache purging
//--------------------------------------------------------------------------

class DummyObj: public ooObj {
public:
    char			dummy[DummySize];
};


// structure for keeping track of a list of base assemblies
// that reference a composite part as either a shared or private member
// these structures are used only during db generation and are 
// not persistent

class BAListElem {
public:
    ooId ba; // reference to a base assembly
    BAListElem*	next;
};

class BAIdList {
private:
    BAListElem* listPtr;
    BAListElem* current;

public:
    BAIdList() {listPtr = NULL; current=NULL;};
    void insert(ooId ba); // insert a new BA ref in list
    ooId next(); // scan list returning all elements
    void  initScan(); // scan list returning all elements
};

