/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"

//////////////////////////////////////////////////////////////////
//
// Reorganization Test #1 - Visit all composite parts, deleting
// and then newly inserting 1/2 of each composite part's atomic
// parts graph.
//
//////////////////////////////////////////////////////////////////


extern char*	types[NumTypes];

extern ooTrans	xact;
extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg1
//
//////////////////////////////////////////////////////////////////


int reorg1(ooHandle(ooDBObj)& databaseH)
{

    // get access to the design library holding composite parts

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocUpdate);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }

    // iterate through all the composite parts, deleting/inserting
    // half of an atomic parts graph worth of parts and counting up
    // the number of atomic parts processed in this way

    int count = 0;
    ooItr(CompositePart) compI;
    for (compI.scan(designLibH, oocUpdate); compI.next(); ) {
        if (debugMode) {
            printf("    In Reorg1, partId = %d:\n", compI->id);
        }
        count += compI->reorg1(designLibH);
    }

    return count;
}



//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg1
//
//////////////////////////////////////////////////////////////////


int CompositePart::reorg1(ooHandle(ooContObj)& designLibH)
{
    if (debugMode) {
        printf("CompositePart::Reorg1(id = %d)\n", id);
    }

    // start by getting pointers to all of the atomic parts, in increasing
    // part id order, in preparation for removing/recreating half of them

    int id, i;
    ooKey key;
    key.type = oocInt32;
    key.offset = ooGetMemberOffset(AtomicPart, id);
    key.size   = ooGetMemberSize(AtomicPart, id);
    key.value  = &id;
    ooHandle(AtomicPart)* atomH = new ooHandle(AtomicPart) [NumAtomicPerComp];
    ooHandle(CompositePart) thisH = ooThis();
    int lowId  = rootPart()->id;
    int highId = lowId + NumAtomicPerComp - 1;

    id = lowId;
    for (i = 0; i < NumAtomicPerComp; i++) {
	atomH[i].lookupObj(designLibH, key);
        if (atomH[i].isValid() == oocFalse) {
            fprintf(stderr, "ERROR: Unable to find atomic part %d\n", id);
   	    xact.abort();
            exit(1);
        }
	id++;
    }

    // now do the deletions of half of the atomic parts, keeping the first
    // half and targeting the second half for simplicity (to avoid the root
    // part and to enable a simple computation to discriminate between parts
    // to be deleted and those to be retained);
    // before doing the deletions, go through and remember all the connections
    // that will be lost when the targeted parts are deleted;
    // after all that, recreate the parts and reconnect them to everything!

    // first compute the range of impact

    int amtToChange, firstToChange, midId;
    amtToChange = NumAtomicPerComp / 2;
    firstToChange = NumAtomicPerComp - amtToChange;
    midId = highId - amtToChange + 1;

    // next record the relevant connections

    PartIdList* toParts = new PartIdList [NumAtomicPerComp];
    for (i = 0; i < NumAtomicPerComp; i++) {
        ooItr(Connection) connI;
        for (atomH[i]->to(connI); connI.next(); ) {
	    if ((atomH[i]->id >= midId) || (connI->to()->id >= midId)) {
		toParts[i].insert(connI->to()->id);
            }
        }
    }

    // now do the actual part deletions

    int count = 0;
    for (i = firstToChange; i < NumAtomicPerComp; i++) {
	if (debugMode) {
            printf("\tReorg1: Deleting atomic part %d\n", atomH[i]->id);
	}
	ooDelete(atomH[i]);
	count++;
    }

    // time to (approximately) recreate the deleted parts

    int32 partDate;
    int typeNo, x, y, length;
    id = midId;
    for (i = firstToChange; i < NumAtomicPerComp; i++) {
	typeNo = (int) (random() % NumTypes);
        partDate = MinAtomicDate + 
	           (int) (random() % (MaxAtomicDate-MinAtomicDate+1));
	x = (int) (random() % XYRange);
	y = (int) (random() % XYRange);
	atomH[i] = ooNewKey(AtomicPart, (id), thisH, key);
	atomH[i]->set_partOf(thisH);
	if (debugMode) {
            printf("\tReorg1: Recreated atomic part %d\n", atomH[i]->id);
	}
	id++;
    }

    // finally, connect all of the parts back up again

    ooHandle(Connection) connH;
    for (i = 0; i < NumAtomicPerComp; i++) {
	while (!toParts[i].empty()) {
	    int to = toParts[i].remove() - lowId;
	    typeNo = (int) (random() % NumTypes);
	    length = (int) (random() % XYRange);
	    connH = new(thisH) Connection (atomH[i], atomH[to]);
	}
    }

    // lastly, clean up and return

    delete [NumAtomicPerComp] atomH;
    delete [NumAtomicPerComp] toParts;
    return count;
}

